
/*************************************************************************************
    Semantic check about HPF distribute directive
    written by Wang Zhenlin
**************************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>
#   include "typedef.h"
#   include "bif.h"
#   include "ll.h"
#   include "symb.h"

#include "defnodes.h"
#include "tags.h"
#include "macro.h"
#include "decl.h"
 
#include "exglobal.h"

extern int is_pure_local_var();

int distribute_attr_smtc(var_list,dist_attr,isDistribute)
PTR_LLND var_list;
PTR_LLND dist_attr;
int      isDistribute; /*if isDistribute=1 ,dist_attr is in distribute stmt or 
			combined stmt else dist_attr is in redistrtibute stmt*/
{
     PTR_SYMB distributee,processors,sym,s;
     int      distributee_dims,dist_format_num,distributed_dims,processors_dims;
     PTR_LLND dist_format,dist_target,cur_llnd;
     int      have_star;
     PTR_TYPE type;
     
     if (dist_attr == LLNULL) /*no dist_format and dist target*/
         return 1;

     dist_format = dist_attr->entry.Template.ll_ptr1;
     dist_target = dist_attr->entry.Template.ll_ptr2;
     
     if (dist_format == LLNULL && dist_target == LLNULL) /*dist_attr is null*/
        return 1;
    
     /* add by LY, check for global symbol */
/*
     check_leaf_sym(dist_format);
     check_leaf_sym(dist_target);
     check_leaf_sym(var_list);
*/

     /* added by LY, to check PURE function */
        if (is_in_pure() && !isDistribute)
                error3(325);

     /*judge if there exists stars before dist_format or dist_target*/     
     have_star = 0;
     if (  (dist_format != LLNULL && dist_format->entry.Template.ll_ptr1 != LLNULL)
             || (dist_target != LLNULL && dist_target->entry.Template.ll_ptr1 != LLNULL))  
         have_star = 1;

     /* star can not be in front of dist_format or dist_target in redistribute stmt*/
     if (have_star && !isDistribute)
        error1(310);

     /*get processors dimensions*/
     /* check dist_target_leaf symbol zheng.gb */
     if ( dist_target == LLNULL )  /*no dist_target*/
        processors_dims = -1;
     else {
	dist_target = dist_target->entry.Template.ll_ptr2;
        if (dist_target == LLNULL)
           processors_dims = -1;
        else {
	   processors = dist_target->entry.Template.symbol;
						/* add by zheng.gb */
	   sym = search_symb_ref(processors);
	   if (sym!=processors) {
		dist_target->entry.Template.symbol=sym;
		if (!symb_in_use(processors)) processors->scope = BFNULL;
		processors = sym;
	   }
	
	   if (processors==SMNULL)
	       processors_dims = -1;
           else {
	   	type=get_base_type(processors->type);
	   	if (type->variant != T_PROCESSORS)
			error3(326,processors->ident);
		else
		if (processors -> type->variant == T_PROCESSORS)
               		processors_dims = 0;
           	else
	       		processors_dims = processors->type->entry.ar_decl.num_dimensions;
	   }

	   /*  check for HPF2.0  zheng.gb 1997/3/13 */
	   if (processors_dims)
	   {
		PTR_LLND sub_list = NODE_OPERAND0(dist_target);
		if (sub_list) {
			int ndim=0;
			for (sub_list;sub_list;sub_list=NODE_OPERAND1(sub_list)) ndim++;
			if (processors_dims != ndim) error3(652,sym->ident);
		}
	   }
	} 
     } 

     if (dist_format == LLNULL ||
	 (dist_format = dist_format->entry.Template.ll_ptr2) == LLNULL ){ 
        dist_format_num = -1;
        distributed_dims = -1;
     }
     else {
        distributed_dims = 0;
        dist_format_num = 0;
	while (dist_format != LLNULL) {
                dist_format_num++;
		if (dist_format->entry.Template.ll_ptr1->variant != STAR_OP)
		{
		   PTR_LLND  op = NODE_OPERAND0(dist_format);
		   switch (op->variant)
		   {
		   case BLOCK_OP:
		   case CYCLIC_OP:
  	           	cur_llnd = NODE_OPERAND0(op);
                   	if (cur_llnd == LLNULL) {  /*no block size or cyclic size)*/ 
		   	}
		   	else {
                       		check_expr_type(cur_llnd);
		       		if (cur_llnd->type->variant != T_INT) {
			 	  	error1(304); 
		       		}
                   	}
			break;
		/* check for HPF 2.0 */
		   case GEN_BLOCK_OP:
		   case INDIRECT_OP:
  	           	cur_llnd = NODE_OPERAND0(op);
                   	if (cur_llnd == LLNULL) {  /*no block size or cyclic size)*/ 
		   	}
		   	else {
                       		check_expr_type(cur_llnd);
		       		if (cur_llnd->type->variant != T_ARRAY)
			 	  	error1(653); 
		       		else if (cur_llnd->type->entry.ar_decl.num_dimensions != 1)
					error1(654);
                   	}
			break;
		   default:
			printf("Unknown dist format:%d. \n",op->variant);
		   }  /* end of switch */
	           distributed_dims++;
                }
                dist_format = NODE_OPERAND1(dist_format);
	 } /*while ; go through every dist_format*/
     }  
     if (processors_dims != -1 && distributed_dims != -1 &&
         distributed_dims != processors_dims)
         error3(306,processors->ident);

     /*check every distributees*/
     while (var_list != LLNULL){ 
	/* add by LY, check for global symbol */
	PTR_LLND  node = var_list->entry.Template.ll_ptr1;
	check_leaf_sym(node);

        /*distributee=node->entry.Template.symbol;*/
        distributee=get_var_name(node);

	/* check for HPF 2.0 */
	/* check if the type of distributee is explicit mapped ! */
	{
	PTR_TYPE base_type=get_base_type(distributee->type);
	if (base_type->variant == T_RECORD)
		if (base_type->entry.re_decl.mapped_flag == 1)
			error3(655,distributee->ident);

	if (node->variant == RECORD_REF) 
		if (isDistribute == 1)   /* in redistribute stmt */
			error3(656,distributee->ident);
		else {
			PTR_LLND part=NODE_OPERAND1(node);
			for ( ; part && NODE_OPERAND1(part) ; part=NODE_OPERAND1(part) )
				if (!is_scalar(NODE_OPERAND0(part)->type))
					error3(657,NODE_SYMB(NODE_OPERAND0(part))->ident);
		}
	}

        /* get dimensions of distributee*/
	/* subobject can not distributed  in HPF 1.0 */
/*
        if (isSubobject(var_list->entry.Template.ll_ptr1)) 
          error3(301,distributee->ident);
       else
*/

       if (node->type->variant == T_ARRAY)
            distributee_dims=node->type->entry.ar_decl.num_dimensions;
       else
            distributee_dims=0;  
      
	/* added by LY, to check PURE function */
	if (is_in_pure() && is_pure_local_var(distributee))
		error3(324, distributee->ident);

       /*if existing star infront of dist_format or dist_target, distributee 
        should have INHERIT attribute*/
/* ?????????????????????????????????????????   zheng.gb add ! befor isD */
/*
       if (isDistribute && have_star && !have_attribute(distributee,INHERIT_BIT)) 
          error3(302,distributee->ident);
*/
       
       /*(re)distributee should not have ALIGN attribute*/
/*
	this will also check in align statement
       if (have_attribute(distributee,ALIGN_BIT)) 
          error3(307,distributee->ident);
*/
       
       /*redistributee should have DYNAMIC attribute*/
       if (!isDistribute && !have_dynamic_attribute(node))
          error3(311,distributee->ident);

       if (    dist_format_num == -1 && processors_dims != -1 
            && processors_dims !=distributee_dims  )  
          error1(303);
       if (dist_format_num != -1 && dist_format_num != distributee_dims)
          error3(305,distributee->ident);

       /*add attr pointer in symboltable*/
       distributee->entry.var_decl.hpfattr = dist_attr;

       var_list = var_list ->entry.Template.ll_ptr2; 
     }
}

/*    check if node have DYNAMIC attr ,  check the rightmost part-name */
int have_dynamic_attribute(node)
PTR_LLND node;
{
	PTR_SYMB name;
	if (node->variant == RECORD_REF)
	{
		while ( NODE_OPERAND1(node) )  node=NODE_OPERAND1(node);
		name=NODE_SYMB(NODE_OPERAND0(node));
	}
	else
        	name=get_var_name(node);
	if (name->attr & DYNAMIC_BIT)
      		return 1;
   	else
      		return 0;
}

int check_data_distribute()
{
    PTR_SYMB blksymb; /*point to the symol of this block*/

    printf("cur variant: %d \n",cur_bfnd->variant);
    blksymb = cur_bfnd->control_parent->entry.Template.symbol;
    if (blksymb->scope == blksymb->thread->scope)
       return 1; /*no variables in this unit*/

    blksymb = blksymb->thread;
    while (blksymb != SMNULL) { 
       if (have_attribute(blksymb,DISTRIBUTE_BIT) || have_attribute(blksymb,ALIGN_BIT)) {
          if (blksymb->entry.var_decl.eq_link.next!=SMNULL){
              if(!have_attribute(blksymb,COVER_BIT)) /*only cover can be mapped*/ 
                  error3(313,blksymb->ident);

              /*only scalar and one rank array can be mapped*/
              if (have_attribute(blksymb,DIMENSION_BIT)) 
                 if (blksymb->type->entry.ar_decl.num_dimensions>1)
                    error3(314,blksymb->ident);
                 else {
                    if (have_attribute(blksymb,DISTRIBUTE_BIT)) {
                       check_equalence_link_dist(blksymb);
                    }
                 } 
          }     
       } 
       blksymb = blksymb->next_symb;
    }   
}

int check_equalence_link_dist(blksymb)
PTR_SYMB blksymb;
{
     PTR_LLND dist_attr;
     int block,size; /*if distribute by BLOCK,block=1. if distribute by CYCLIC block=0*/  
     PTR_SYMB distributee,processors,equal_var;
     int      processors_dims,processors_num,distributee_len,dims;
     PTR_LLND dist_format,dist_target,cur_llnd;
     int      variant,up[7],down[7],tag[7];
 
     /*get dimension of processors*/
     dist_attr = blksymb->entry.var_decl.hpfattr; 
     if (dist_attr == LLNULL) /*no dist_format and dist target*/
         return 1;

     dist_format = dist_attr->entry.Template.ll_ptr1;
     dist_target = dist_attr->entry.Template.ll_ptr2;
     
     if (dist_format == LLNULL && dist_target == LLNULL)
        return 1;

     /*get processors dimensions*/
     if ( dist_target == LLNULL )  /*no dist_target*/
        processors_num = -1;
     else {
	dist_target = dist_target->entry.Template.ll_ptr2;
        if (dist_target == LLNULL)
           processors_num = -1;
        else {
	   processors = dist_target->entry.Template.symbol;
	
	   if (processors==SMNULL)
	       processors_num = -1;
           else if (processors -> type->variant == T_PROCESSORS)
               processors_num = 1;
           else {
	       processors_dims = processors->type->entry.ar_decl.num_dimensions; 
               if (processors_dims >1) /*processors sholuld be scalar 
                                       or rank_one array here*/
                   return 0;
               else {
                  get_bound_of_array(processors->type,down,up,tag,&processors_dims);
                  if (tag[0]==1) 
                    processors_num = up[0]-down[0];
                  else
                    processors_num = -1;
               }           
           }
        } 
     } 

     /*get dimension extend of ditributee*/
     distributee_len = type_size(distributee->type);

     if (dist_format == LLNULL ||
	 (dist_format = dist_format->entry.Template.ll_ptr2) == LLNULL ){ 
         return 1;
     }
     else {
	variant = dist_format->entry.Template.ll_ptr1->variant; 
	if (variant != STAR_OP){
            cur_llnd = dist_format->entry.Template.ll_ptr1->entry.Template.ll_ptr1;
            /*get size*/
            if (cur_llnd == LLNULL) /*no block size or cyclic size)*/ {
                 if (processors_num <= 1 || distributee_len == -1) 
                     return 1;

                 if (dist_format->entry.Template.ll_ptr1->variant == BLOCK_OP)
                     size = CD(distributee_len,processors_num);
                 else 
                     size =1;  
	    }
            else {
                   if (comput_const_expr(cur_llnd)==0) 
                      return 1;
                   else
                     size = cur_llnd ->const_val;
                 }
            size = size * type_size(distributee->type);
            
            equal_var = distributee->entry.var_decl.eq_link.next;
            if (distributee->entry.var_decl.eq_link.offset !=0 ) {
               error1(1);  /*internal error*/
            }

            while ( equal_var != distributee)  {
                check_each_equal_var(equal_var,distributee,size);
                equal_var = equal_var->entry.var_decl.eq_link.next;
            }
        } 
    } /*dist_format not null*/
}

int get_bound_of_array(type,down,up,tag,dims)
PTR_TYPE type;
int down[],up[],tag[],*dims;
{  
   int i;  /*tag[i]=0,no up and down bound*/
           /*tag[i]=1, existing both const up and down bound*/
           /*tag[i]=2, no const down bound*/
           /*tag[i]=3, no const up bound*/

   int dimcount=0;
   PTR_LLND  ranges,downllnd,upllnd;/*dim ranges*/

   for (i=0;i<7;i++)
     tag[i]=1;

   *dims = type->entry.ar_decl.num_dimensions;
   ranges = type->entry.ar_decl.ranges;
   while (ranges !=LLNULL) {
       downllnd = ranges->entry.Template.ll_ptr1->entry.Template.ll_ptr1->entry.Template.ll_ptr1;     
       upllnd = ranges->entry.Template.ll_ptr1->entry.Template.ll_ptr1->entry.Template.ll_ptr2;     
       if (downllnd != LLNULL || upllnd != LLNULL){
         if (downllnd != LLNULL) {
            if (comput_const_expr(downllnd,1))
               down[dimcount] = downllnd->const_val;
            else 
               tag[dimcount]=2;
         }
         else 
             down[dimcount]=1;

         if (upllnd != LLNULL) {
            if (comput_const_expr(upllnd,1))
               up[dimcount] = upllnd->const_val;
            else {
               if (tag[dimcount] ==2)
                   tag[dimcount] = 0;
               else
                   tag[dimcount]=3;
            }
         }
         else {
               if (tag[dimcount] ==2)
                   tag[dimcount] = 0;
               else
                   tag[dimcount]=3;
         }

       }
       else
          tag[dimcount]=0; 
       dimcount++;
       ranges = ranges->entry.Template.ll_ptr2;
   }

}

int CD(a,b)
int a,b;
{
   return ( (a+b-1)/b );
}


int check_each_equal_var(symb,dis_symb,block_size)
PTR_SYMB symb,dis_symb;
int block_size;
{
   int i,offset,size_l,size_r;
   int symb_size,element_size;
   int up[7],down[7],tag[7],dims;
   
   if (have_attribute(symb,DIMENSION_BIT)){ /* is array*/
      element_size = type_size(symb->type->entry.ar_decl.base_type);  
      symb_size *= type_size(symb->type)*element_size;
   }
   else {
      symb_size = element_size = type_size(symb->type); 
   }
   get_bound_of_array(symb,down,up,tag);  
   
   offset = symb->entry.var_decl.eq_link.offset;
   if (offset < 0) {
      error1(1);
      return 0;
   }
   size_l = offset % block_size;
   size_r = (symb_size-(block_size-size_l)) % block_size;
   if (  size_l % element_size != 0 
      || size_r % element_size != 0
      || (size_l+size_r<symb_size && block_size%element_size!=0)) {
       error4(315,dis_symb->ident,symb->ident);
   }   
}
